/*----------------------------------------------------

   LZFR.EXE

   LZIL.IL pt@CkWJc[ Ver.1.00
   
   Based on source listed by Haruhiko Okumura.
     http://www.matsusaka-u.ac.jp/~okumura/

   For WonderWitch tuned by Renya Iizuka. 2003/11/15
     http://www.netlaputa.ne.jp/~pastelsi/

-----------------------------------------------------*/
typedef	unsigned short	WORD;
typedef unsigned char	BYTE;
typedef unsigned long	DWORD;

#include <stdio.h>
#include <stdlib.h>
#include "resource.h"
#include "lzfr.h"

#define	PAL_SIZE	(16 * sizeof(short))

#define N   4096  /* obt@̑傫 */
#define F     18  /* Œv */
#define	NIL    N  /* ؂̖[ */


FILE			*infile, *outfile;      /* ̓t@C, o̓t@C */
unsigned long	outcount = 0;  /* o̓oCgJE^ */
unsigned char	text[N+F-1]; /* eLXgpobt@ */
int				dad[N+1], lson[N+1], rson[N+257];  /*  */

/*-------------------------------------------------------------*/
long GetNumber(unsigned char *p){

	int		a;
	long	ret = 0;

	while( a = *p++ ){
		if( a<'0' || a>'9' ){
			return -1;
		}
		ret *= 10;
		ret += a - '0';
	}

	return ret;

}

/*-------------------------------------------------------------*/
void error(char *message)  /* bZ[W\I */
{
	fprintf(stderr, "\n%s\n", message);
	exit(EXIT_FAILURE);
}

/*-------------------------------------------------------------*/
void init_tree(void)  /* ؂̏ */
{
	int i;

	for (i = N + 1; i <= N + 256; i++) rson[i] = NIL;
	for (i = 0; i < N; i++) dad[i] = NIL;
}

int matchpos, matchlen;  /* Œvʒu, v */

/*-------------------------------------------------------------*/
void insert_node(int r)  /*  r ؂ɑ} */
{
	int i, p, cmp;
	unsigned char *key;

	cmp = 1;  key = &text[r];  p = N + 1 + key[0];
	rson[r] = lson[r] = NIL;  matchlen = 0;
	for ( ; ; ) {
		if (cmp >= 0) {
			if (rson[p] != NIL) p = rson[p];
			else {  rson[p] = r;  dad[r] = p;  return;  }
		} else {
			if (lson[p] != NIL) p = lson[p];
			else {  lson[p] = r;  dad[r] = p;  return;  }
		}
		for (i = 1; i < F; i++)
			if ((cmp = key[i] - text[p + i]) != 0)  break;
		if (i > matchlen) {
			matchpos = p;
			if ((matchlen = i) >= F)  break;
		}
	}
	dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
	dad[lson[p]] = r;  dad[rson[p]] = r;
	if (rson[dad[p]] == p) rson[dad[p]] = r;
	else                   lson[dad[p]] = r;
	dad[p] = NIL;  /* p O */
}

/*-------------------------------------------------------------*/
void delete_node(int p)  /*  p ؂ */
{
	int  q;

	if (dad[p] == NIL) return;  /* Ȃ */
	if (rson[p] == NIL) q = lson[p];
	else if (lson[p] == NIL) q = rson[p];
	else {
		q = lson[p];
		if (rson[q] != NIL) {
			do {  q = rson[q];  } while (rson[q] != NIL);
			rson[dad[q]] = lson[q];  dad[lson[q]] = dad[q];
			lson[q] = lson[p];  dad[lson[p]] = q;
		}
		rson[q] = rson[p];  dad[rson[p]] = q;
	}
	dad[q] = dad[p];
	if (rson[dad[p]] == p) rson[dad[p]] = q;
	else                   lson[dad[p]] = q;
	dad[p] = NIL;
}

/*-------------------------------------------------------------*/
unsigned long encode(void)  /* k */
{
	int i, c, len, r, s, lastmatchlen, codeptr;
	unsigned char code[17], mask;
	unsigned long int incount = 0, printcount = 0, cr;

	init_tree();  /* ؂ */
	code[0] = 0;  codeptr = mask = 1;
	s = 0;  r = N - F;
	for (i = s; i < r; i++) text[i] = 0;  /* obt@ */
	for (len = 0; len < F ; len++) {
		c = getc(infile);  if (c == EOF) break;
		text[r + len] = c;
	}
	incount = len;  if (incount == 0) return 0;
	for (i = 1; i <= F; i++) insert_node(r - i);
	insert_node(r);
	do {
		if (matchlen > len) matchlen = len;
		if (matchlen < 3) {
			matchlen = 1;  code[0] |= mask;  code[codeptr++] = text[r];
		} else {
			code[codeptr++] = (unsigned char) matchpos;
			code[codeptr++] = (unsigned char)
				(((matchpos >> 4) & 0xf0) | (matchlen - 3));
		}
		if ((mask <<= 1) == 0) {
			for (i = 0; i < codeptr; i++) putc(code[i], outfile);
			outcount += codeptr;
			code[0] = 0;  codeptr = mask = 1;
		}
		lastmatchlen = matchlen;
		for (i = 0; i < lastmatchlen; i++) {
			c = getc(infile);  if (c == EOF) break;
			delete_node(s);  text[s] = c;
			if (s < F - 1) text[s + N] = c;
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			insert_node(r);
		}
		if ((incount += i) > printcount) {
			printf("%12lu\r", incount);  printcount += 1024;
		}
		while (i++ < lastmatchlen) {
			delete_node(s);
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			if (--len) insert_node(r);
		}
	} while (len > 0);
	if (codeptr > 1) {
		for (i = 0; i < codeptr; i++) putc(code[i], outfile);
		outcount += codeptr;
	}
	printf(" : %lu oCg\n", incount);  /* ʕ */
	printf("o : %lu oCg\n", outcount);
	if (incount != 0) {  /* k߂ĕ */
		cr = (outcount*100) / incount;
		printf("k : %d%%\n", cr);
	}

	return outcount;

}

/*-------------------------------------------------------------*/
void decode(unsigned long int size)  /*  */
{
	int i, j, k, r, c;
	unsigned int flags;

	for (i = 0; i < N - F; i++) text[i] = 0;
	r = N - F;  flags = 0;
	for ( ; ; ) {
		if (((flags >>= 1) & 256) == 0) {
			if ((c = getc(infile)) == EOF) break;
			flags = c | 0xff00;
		}
		if (flags & 1) {
			if ((c = getc(infile)) == EOF) break;
			putc(c, outfile);  text[r++] = c;  r &= (N - 1);
		} else {
			if ((i = getc(infile)) == EOF) break;
			if ((j = getc(infile)) == EOF) break;
			i |= ((j & 0xf0) << 4);  j = (j & 0x0f) + 2;
			for (k = 0; k <= j; k++) {
				c = text[(i + k) & (N - 1)];  putc(c, outfile);
				text[r++] = c;  r &= (N - 1);
			}
		}
	}
	printf("WJoCg : %d oCg\n", size);
}

/*-------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	int					c;
	unsigned short		magic  = 'F' + ('R'<<8);	/* 'FR' */
	unsigned short		typeLZ = 'L' + ('Z'<<8);	/* 'LZ' */
	unsigned short		typeFR = 'F' + ('R'<<8);	/* 'LZ' */
	unsigned long		size16p;
	unsigned short		id = 0;
	unsigned short		typeDT = 'D'+('T'<<8);		/* 'DT' */
	unsigned short		typeBM = 'B'+('M'<<8);		/* 'BM' */
	unsigned long		size;									/* ̃oCg */
	unsigned char		read_header[ sizeof(RES_LZ) ];			/* wb_ʗp */
	res_image_t			*pRes   = (res_image_t *)read_header;
	RES_LZ				*pResLZ = (RES_LZ *)read_header;
	long				num;


	if (argc != 4 || ((c = *argv[1]) != 'E' && c != 'e'
								&& c != 'F' && c != 'f'
								&& c != 'D' && c != 'd'
								&& c != 'C' && c != 'c') )
	{
		printf("LZFR : WonderWitch LZIL pt@CkWJc[łB\n");
		printf("Copyright(C)2003 EXCEED. / Based on source coded by Haruhiko Okumura.\n");
		printf("gp@ : LZFR XCb`[oIDԍ] ̓t@C o̓t@C\n\n");
		printf("<< XCb`̐ >>\n"); 
		printf("  c[l]  : FRBM `̃wb_؂ LZDT `ňk܂B\n"); 
		printf("  e[l]  : wb_ʈk܂B\n"); 
		printf("  f[l]  : wb_ʂɋk܂B\n"); 
		printf("  d[l]  : kWJ܂B\n"); 
		printf(" [l]ȗƁA0 w肳ꂽ̂ƌȂ܂B\n");
		return EXIT_FAILURE;
	}

	if ((infile  = fopen(argv[2], "rb")) == NULL){
		error("G[ : ̓t@CJ܂łB");
	}

	if ((outfile = fopen(argv[3], "wb")) == NULL){
		error("G[ : o̓t@C܂B");
	}

	num = GetNumber( (argv[1]) + 1 );

	if( num < 0 || num > 65535){
		error("G[ : w ID ̒lُłB");
	}

	id = (unsigned short) num;

	if ( c=='D' || c=='d' ){

		/* ----------  ---------- */

		/* wb_ʂ */
		rewind(infile);
		fread(&read_header, sizeof(char), sizeof(RES_LZ), infile);

		if(	pRes->header.magic != magic || pRes->header.type != typeLZ ){
			error("G[ : ̓t@ĆALZIL pt@Cł͂܂B");
		}

		if(	pResLZ->LZ.type2 == typeBM ) {

			/* BM WJ */
			size    = pResLZ->TYPE.BM.size;
			size16p = size + sizeof(res_image_t) + PAL_SIZE;
			size16p = (size16p + 15) / 16;

			fwrite( &magic,   sizeof(pResLZ->LZ.magic),   1, outfile );
			fwrite( &typeBM,  sizeof(pResLZ->LZ.type),    1, outfile );
			fwrite( &size16p, sizeof(pResLZ->LZ.size16p), 1, outfile );
			fwrite( &id,      sizeof(pResLZ->LZ.id),      1, outfile );

			fwrite( &pResLZ->TYPE.BM.width,    sizeof(pResLZ->TYPE.BM.width),    1, outfile );
			fwrite( &pResLZ->TYPE.BM.height,   sizeof(pResLZ->TYPE.BM.height),   1, outfile );
			fwrite( &pResLZ->TYPE.BM.format,   sizeof(pResLZ->TYPE.BM.format),   1, outfile );
			fwrite( &pResLZ->TYPE.BM.reserve,  sizeof(pResLZ->TYPE.BM.reserve),  1, outfile );
			fwrite( &pResLZ->TYPE.BM.clrtable, sizeof(pResLZ->TYPE.BM.clrtable), 1, outfile );
			fwrite( (BYTE *)&pResLZ->TYPE.BM.clrtable + sizeof(pResLZ->TYPE.BM.clrtable), sizeof(short), 16, outfile);

			rewind(infile);
			fseek( infile, (long)(sizeof(RES_COMMON_HEADER) + sizeof(TYPE_BM)), SEEK_SET );

			decode(size);

		} else {

			/* DT WJ */
			size = pResLZ->TYPE.DT.size;
			rewind(infile);
			fseek(infile, (long)(sizeof(RES_COMMON_HEADER) + sizeof(TYPE_DT)), SEEK_SET);

			decode(size); 

		}

	} else {

		fseek(infile, 0L, SEEK_END);	/* infile ̖T */
		size = ftell(infile);			/* infile ̃oCg */
		if(size>65536){
			error("G[ : ̓t@C̃TCY傫̂ŁAk͂ł܂B");
		}

		/* wb_ʂ */
		rewind(infile);
		fread(&read_header, sizeof(char), sizeof(res_image_t) + PAL_SIZE, infile);
		rewind(infile);

		if( c=='C' || c=='c' ){
			if(pRes->header.type != typeBM){
				error ("G[ : ̓t@CAFRBM `ł͂܂B");
			}
			size -= 48;
			fseek(infile, 48L, SEEK_SET);
		}

		if(  (c=='C' || c=='c') || (c=='F' || c=='f')  ||  pRes->header.type != typeBM  ){

			/* FRDT */
			fwrite(&magic, sizeof(short), 1, outfile);
			fwrite(&typeLZ, sizeof(short), 1, outfile);
			fwrite(&size16p, sizeof(short), 1, outfile);	/* _~[ */
			fwrite(&id, sizeof(short), 1, outfile);

			fwrite(&typeDT, sizeof(short), 1, outfile);
			fwrite(&size, sizeof(short), 1, outfile);

			size16p = encode();

			size16p += sizeof(RES_COMMON_HEADER) + sizeof(TYPE_DT);		/* wb_TCYZ */
			size16p = (size16p + 15) / 16;

			fseek(outfile, 4L, SEEK_SET);
			fwrite(&size16p, sizeof(short), 1, outfile);	/* Ȓl߂ */

		} else {

			/* FRBM */

			size -= sizeof(res_image_t) + PAL_SIZE;		/* wb_͈kȂ */

			fwrite(&magic, sizeof(short), 1, outfile);
			fwrite(&typeLZ, sizeof(short), 1, outfile);
			fwrite(&size16p, sizeof(short), 1, outfile);	/* _~[ */
			fwrite(&id, sizeof(short), 1, outfile);

			fwrite(&typeBM, sizeof(short), 1, outfile);
			fwrite(&size, sizeof(short), 1, outfile);

			fwrite(&pRes->width, sizeof(pRes->width), 1, outfile);
			fwrite(&pRes->height, sizeof(pRes->height), 1, outfile);
			fwrite(&pRes->format, sizeof(pRes->format), 1, outfile);
			fwrite(&pRes->reserve, sizeof(pRes->reserve), 1, outfile);
			fwrite(&pRes->clrtable, sizeof(pRes->clrtable), 1, outfile);
			fwrite((BYTE *)&pRes->clrtable + sizeof(pRes->clrtable), sizeof(short), 16, outfile);
			
			fseek(infile, (long)(sizeof(res_image_t) + PAL_SIZE), SEEK_SET);

			size16p = encode();

			size16p += sizeof(RES_LZ);
			size16p = (size16p + 15) / 16;

			fseek(outfile, 4L, SEEK_SET);
			fwrite(&size16p, sizeof(short), 1, outfile);	/* Ȓl߂ */

		}

	} 

	fclose(infile);
	fclose(outfile);

	printf("܂B\n");

	return EXIT_SUCCESS;

}

